@@ -3,12 +3,12 @@
  * date.c
  *	  implements DATE and TIME data types specified in SQL-92 standard
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994-5, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/adt/date.c,v 1.64.2.4 2005/04/23 22:54:23 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/date.c,v 1.73.2.1 2002/11/21 23:31:37 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -27,7 +27,19 @@
 #include "utils/nabstime.h"
 #include "utils/timestamp.h"
 
+/*
+ * gcc's -ffast-math switch breaks routines that expect exact results from
+ * expressions like timeval / 3600, where timeval is double.
+ */
+#ifdef __FAST_MATH__
+#error -ffast-math is known to break this code
+#endif
 
+
+static int	time2tm(TimeADT time, struct tm * tm, fsec_t *fsec);
+static int	timetz2tm(TimeTzADT *time, struct tm * tm, fsec_t *fsec, int *tzp);
+static int	tm2time(struct tm * tm, fsec_t fsec, TimeADT *result);
+static int	tm2timetz(struct tm * tm, fsec_t fsec, int tz, TimeTzADT *result);
 static void AdjustTimeForTypmod(TimeADT *time, int32 typmod);
 
 /*****************************************************************************
@@ -43,7 +55,7 @@ date_in(PG_FUNCTION_ARGS)
 {
 	char	   *str = PG_GETARG_CSTRING(0);
 	DateADT		date;
-	double		fsec;
+	fsec_t		fsec;
 	struct tm	tt,
 			   *tm = &tt;
 	int			tzp;
@@ -68,7 +80,7 @@ date_in(PG_FUNCTION_ARGS)
 		case DTK_CURRENT:
 			elog(ERROR, "Date CURRENT no longer supported"
 				 "\n\tdate_in() internal coding error");
-			GetCurrentTime(tm);
+			GetCurrentDateTime(tm);
 			break;
 
 		case DTK_EPOCH:
@@ -224,6 +236,60 @@ date_mii(PG_FUNCTION_ARGS)
 	PG_RETURN_DATEADT(dateVal - days);
 }
 
+#if NOT_USED
+/* date_pl_interval() and date_mi_interval() are probably
+ * better implmented by converting the input date
+ * to timestamp without time zone. So that is what we do
+ * in pg_proc.h - thomas 2002-03-11
+ */
+
+/* Add an interval to a date, giving a new date.
+ * Must handle both positive and negative intervals.
+ */
+Datum
+date_pl_interval(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Interval   *span = PG_GETARG_INTERVAL_P(1);
+	struct tm	tt,
+			   *tm = &tt;
+
+	if (span->month != 0)
+	{
+		j2date((dateVal + date2j(2000, 1, 1)), &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
+		tm->tm_mon += span->month;
+		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+	}
+	if (span->time != 0)
+		dateVal += (span->time / 86400e0);
+
+	PG_RETURN_DATEADT(dateVal);
+}
+
+/* Subtract an interval from a date, giving a new date.
+ * Must handle both positive and negative intervals.
+ */
+Datum
+date_mi_interval(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Interval   *span = PG_GETARG_INTERVAL_P(1);
+	struct tm	tt,
+			   *tm = &tt;
+
+	if (span->month != 0)
+	{
+		j2date((dateVal + date2j(2000, 1, 1)), &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
+		tm->tm_mon -= span->month;
+		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+	}
+	if (span->time != 0)
+		dateVal -= (span->time / 86400e0);
+
+	PG_RETURN_DATEADT(dateVal);
+}
+#endif
+
 /* date_timestamp()
  * Convert date to timestamp data type.
  */
@@ -233,8 +299,13 @@ date_timestamp(PG_FUNCTION_ARGS)
 	DateADT		dateVal = PG_GETARG_DATEADT(0);
 	Timestamp	result;
 
+#ifdef HAVE_INT64_TIMESTAMP
+	/* date is days since 2000, timestamp is microseconds since same... */
+	result = dateVal * INT64CONST(86400000000);
+#else
 	/* date is days since 2000, timestamp is seconds since same... */
 	result = dateVal * 86400.0;
+#endif
 
 	PG_RETURN_TIMESTAMP(result);
 }
@@ -248,17 +319,17 @@ timestamp_date(PG_FUNCTION_ARGS)
 {
 	Timestamp	timestamp = PG_GETARG_TIMESTAMP(0);
 	DateADT		result;
-	struct tm	tt,
-			   *tm = &tt;
-	double		fsec;
 
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
-	if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL) != 0)
-		elog(ERROR, "Unable to convert timestamp to date");
-
-	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+#ifdef HAVE_INT64_TIMESTAMP
+	/* Microseconds to days */
+	result = (timestamp / INT64CONST(86400000000));
+#else
+	/* Seconds to days */
+	result = (timestamp / 86400.0);
+#endif
 
 	PG_RETURN_DATEADT(result);
 }
@@ -287,12 +358,21 @@ date_timestamptz(PG_FUNCTION_ARGS)
 		tm->tm_sec = 0;
 		tz = DetermineLocalTimeZone(tm);
 
+#ifdef HAVE_INT64_TIMESTAMP
+		result = (dateVal * INT64CONST(86400000000))
+			+ (tz * INT64CONST(1000000));
+#else
 		result = dateVal * 86400.0 + tz;
+#endif
 	}
 	else
 	{
 		/* Outside of range for timezone support, so assume UTC */
+#ifdef HAVE_INT64_TIMESTAMP
+		result = (dateVal * INT64CONST(86400000000));
+#else
 		result = dateVal * 86400.0;
+#endif
 	}
 
 	PG_RETURN_TIMESTAMP(result);
@@ -309,7 +389,7 @@ timestamptz_date(PG_FUNCTION_ARGS)
 	DateADT		result;
 	struct tm	tt,
 			   *tm = &tt;
-	double		fsec;
+	fsec_t		fsec;
 	int			tz;
 	char	   *tzn;
 
@@ -428,7 +508,7 @@ time_in(PG_FUNCTION_ARGS)
 #endif
 	int32		typmod = PG_GETARG_INT32(2);
 	TimeADT		result;
-	double		fsec;
+	fsec_t		fsec;
 	struct tm	tt,
 			   *tm = &tt;
 	int			nf;
@@ -444,13 +524,56 @@ time_in(PG_FUNCTION_ARGS)
 	 || (DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, NULL) != 0))
 		elog(ERROR, "Bad time external representation '%s'", str);
 
-	result = ((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec + fsec);
-
+	tm2time(tm, fsec, &result);
 	AdjustTimeForTypmod(&result, typmod);
 
 	PG_RETURN_TIMEADT(result);
 }
 
+/* tm2time()
+ * Convert a tm structure to a time data type.
+ */
+static int
+tm2time(struct tm * tm, fsec_t fsec, TimeADT *result)
+{
+#ifdef HAVE_INT64_TIMESTAMP
+	*result = ((((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec)
+				* INT64CONST(1000000)) + fsec);
+#else
+	*result = ((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec + fsec);
+#endif
+	return 0;
+}
+
+/* time2tm()
+ * Convert time data type to POSIX time structure.
+ * For dates within the system-supported time_t range, convert to the
+ *	local time zone. If out of this range, leave as GMT. - tgl 97/05/27
+ */
+static int
+time2tm(TimeADT time, struct tm * tm, fsec_t *fsec)
+{
+#ifdef HAVE_INT64_TIMESTAMP
+	tm->tm_hour = (time / INT64CONST(3600000000));
+	time -= (tm->tm_hour * INT64CONST(3600000000));
+	tm->tm_min = (time / INT64CONST(60000000));
+	time -= (tm->tm_min * INT64CONST(60000000));
+	tm->tm_sec = (time / INT64CONST(1000000));
+	time -= (tm->tm_sec * INT64CONST(1000000));
+	*fsec = time;
+#else
+	double		trem;
+
+	trem = time;
+	TMODULO(trem, tm->tm_hour, 3600e0);
+	TMODULO(trem, tm->tm_min, 60e0);
+	TMODULO(trem, tm->tm_sec, 1e0);
+	*fsec = trem;
+#endif
+
+	return 0;
+}
+
 Datum
 time_out(PG_FUNCTION_ARGS)
 {
@@ -458,16 +581,10 @@ time_out(PG_FUNCTION_ARGS)
 	char	   *result;
 	struct tm	tt,
 			   *tm = &tt;
-	double		fsec;
-	double		trem;
+	fsec_t		fsec;
 	char		buf[MAXDATELEN + 1];
 
-	trem = time;
-	TMODULO(trem, tm->tm_hour, 3600e0);
-	TMODULO(trem, tm->tm_min, 60e0);
-	TMODULO(trem, tm->tm_sec, 1e0);
-	fsec = trem;
-
+	time2tm(time, tm, &fsec);
 	EncodeTimeOnly(tm, fsec, NULL, DateStyle, buf);
 
 	result = pstrdup(buf);
@@ -491,24 +608,90 @@ time_scale(PG_FUNCTION_ARGS)
 	PG_RETURN_TIMEADT(result);
 }
 
+/* AdjustTimeForTypmod()
+ * Force the precision of the time value to a specified value.
+ * Uses *exactly* the same code as in AdjustTimestampForTypemod()
+ * but we make a separate copy because those types do not
+ * have a fundamental tie together but rather a coincidence of
+ * implementation. - thomas
+ */
 static void
 AdjustTimeForTypmod(TimeADT *time, int32 typmod)
 {
-	if ((typmod >= 0) && (typmod <= 13))
+#ifdef HAVE_INT64_TIMESTAMP
+	static const int64 TimeScales[MAX_TIMESTAMP_PRECISION + 1] = {
+		INT64CONST(1000000),
+		INT64CONST(100000),
+		INT64CONST(10000),
+		INT64CONST(1000),
+		INT64CONST(100),
+		INT64CONST(10),
+		INT64CONST(1)
+	};
+
+	static const int64 TimeOffsets[MAX_TIMESTAMP_PRECISION + 1] = {
+		INT64CONST(-500000),
+		INT64CONST(-50000),
+		INT64CONST(-5000),
+		INT64CONST(-500),
+		INT64CONST(-50),
+		INT64CONST(-5),
+		INT64CONST(0)
+	};
+
+#else
+	static const double TimeScales[MAX_TIMESTAMP_PRECISION + 1] = {
+		1,
+		10,
+		100,
+		1000,
+		10000,
+		100000,
+		1000000
+	};
+
+	static const double TimeOffsets[MAX_TIMESTAMP_PRECISION + 1] = {
+		0.5,
+		0.05,
+		0.005,
+		0.0005,
+		0.00005,
+		0.000005,
+		0.0000005
+	};
+#endif
+
+	if ((typmod >= 0) && (typmod <= MAX_TIME_PRECISION))
 	{
-		static double TimeScale = 1;
-		static int32 TimeTypmod = 0;
-
-		if (typmod != TimeTypmod)
+#ifdef HAVE_INT64_TIMESTAMP
+		/* we have different truncation behavior depending on sign */
+		if (*time >= INT64CONST(0))
 		{
-			TimeScale = pow(10.0, typmod);
-			TimeTypmod = typmod;
+			*time = ((*time / TimeScales[typmod])
+					 * TimeScales[typmod]);
 		}
-
-		*time = (rint(((double) *time) * TimeScale) / TimeScale);
-
-		if (*time >= 86400)
-			*time -= 86400;
+		else
+		{
+			*time = (((*time + TimeOffsets[typmod]) / TimeScales[typmod])
+					 * TimeScales[typmod]);
+		}
+#else
+		/* we have different truncation behavior depending on sign */
+		if (*time >= 0)
+		{
+			*time = (rint(((double) *time) * TimeScales[typmod])
+					 / TimeScales[typmod]);
+		}
+		else
+		{
+			/*
+			 * Scale and truncate first, then add to help the rounding
+			 * behavior
+			 */
+			*time = (rint((((double) *time) * TimeScales[typmod]) + TimeOffsets[typmod])
+					 / TimeScales[typmod]);
+		}
+#endif
 	}
 
 	return;
@@ -736,15 +919,60 @@ timestamp_time(PG_FUNCTION_ARGS)
 	TimeADT		result;
 	struct tm	tt,
 			   *tm = &tt;
-	double		fsec;
+	fsec_t		fsec;
 
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
 	if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL) != 0)
-		elog(ERROR, "Unable to convert timestamp to date");
+		elog(ERROR, "Unable to convert timestamp to time");
 
+#ifdef HAVE_INT64_TIMESTAMP
+
+	/*
+	 * Could also do this with time = (timestamp / 86400000000 *
+	 * 86400000000) - timestamp;
+	 */
+	result = ((((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec)
+			   * INT64CONST(1000000)) + fsec);
+#else
 	result = ((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec + fsec);
+#endif
+
+	PG_RETURN_TIMEADT(result);
+}
+
+/* timestamptz_time()
+ * Convert timestamptz to time data type.
+ */
+Datum
+timestamptz_time(PG_FUNCTION_ARGS)
+{
+	TimestampTz timestamp = PG_GETARG_TIMESTAMP(0);
+	TimeADT		result;
+	struct tm	tt,
+			   *tm = &tt;
+	int			tz;
+	fsec_t		fsec;
+	char	   *tzn;
+
+	if (TIMESTAMP_NOT_FINITE(timestamp))
+		PG_RETURN_NULL();
+
+	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
+		elog(ERROR, "Unable to convert timestamptz to time");
+
+#ifdef HAVE_INT64_TIMESTAMP
+
+	/*
+	 * Could also do this with time = (timestamp / 86400000000 *
+	 * 86400000000) - timestamp;
+	 */
+	result = ((((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec)
+			   * INT64CONST(1000000)) + fsec);
+#else
+	result = ((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec + fsec);
+#endif
 
 	PG_RETURN_TIMEADT(result);
 }
@@ -791,10 +1019,18 @@ interval_time(PG_FUNCTION_ARGS)
 {
 	Interval   *span = PG_GETARG_INTERVAL_P(0);
 	TimeADT		result;
+
+#ifdef HAVE_INT64_TIMESTAMP
+	result = span->time;
+	if ((result >= INT64CONST(86400000000))
+		|| (result <= INT64CONST(-86400000000)))
+		result -= (result / INT64CONST(1000000) * INT64CONST(1000000));
+#else
 	Interval	span1;
 
 	result = span->time;
 	TMODULO(result, span1.time, 86400e0);
+#endif
 
 	PG_RETURN_TIMEADT(result);
 }
@@ -811,7 +1047,7 @@ time_mi_time(PG_FUNCTION_ARGS)
 
 	result = (Interval *) palloc(sizeof(Interval));
 
-	result->time = time1 - time2;
+	result->time = (time1 - time2);
 	result->month = 0;
 
 	PG_RETURN_INTERVAL_P(result);
@@ -826,12 +1062,20 @@ time_pl_interval(PG_FUNCTION_ARGS)
 	TimeADT		time = PG_GETARG_TIMEADT(0);
 	Interval   *span = PG_GETARG_INTERVAL_P(1);
 	TimeADT		result;
+
+#ifdef HAVE_INT64_TIMESTAMP
+	result = (time + span->time);
+	result -= (result / INT64CONST(86400000000) * INT64CONST(86400000000));
+	if (result < INT64CONST(0))
+		result += INT64CONST(86400000000);
+#else
 	TimeADT		time1;
 
 	result = (time + span->time);
 	TMODULO(result, time1, 86400e0);
 	if (result < 0)
 		result += 86400;
+#endif
 
 	PG_RETURN_TIMEADT(result);
 }
@@ -845,12 +1089,20 @@ time_mi_interval(PG_FUNCTION_ARGS)
 	TimeADT		time = PG_GETARG_TIMEADT(0);
 	Interval   *span = PG_GETARG_INTERVAL_P(1);
 	TimeADT		result;
+
+#ifdef HAVE_INT64_TIMESTAMP
+	result = (time - span->time);
+	result -= (result / INT64CONST(86400000000) * INT64CONST(86400000000));
+	if (result < INT64CONST(0))
+		result += INT64CONST(86400000000);
+#else
 	TimeADT		time1;
 
 	result = (time - span->time);
 	TMODULO(result, time1, 86400e0);
 	if (result < 0)
 		result += 86400;
+#endif
 
 	PG_RETURN_TIMEADT(result);
 }
@@ -924,11 +1176,137 @@ text_time(PG_FUNCTION_ARGS)
 							   Int32GetDatum(-1));
 }
 
+/* time_part()
+ * Extract specified field from time type.
+ */
+Datum
+time_part(PG_FUNCTION_ARGS)
+{
+	text	   *units = PG_GETARG_TEXT_P(0);
+	TimeADT		time = PG_GETARG_TIMEADT(1);
+	float8		result;
+	int			type,
+				val;
+	int			i;
+	char	   *up,
+			   *lp,
+				lowunits[MAXDATELEN + 1];
+
+	if (VARSIZE(units) - VARHDRSZ > MAXDATELEN)
+		elog(ERROR, "TIME units '%s' not recognized",
+			 DatumGetCString(DirectFunctionCall1(textout,
+											   PointerGetDatum(units))));
+	up = VARDATA(units);
+	lp = lowunits;
+	for (i = 0; i < (VARSIZE(units) - VARHDRSZ); i++)
+		*lp++ = tolower((unsigned char) *up++);
+	*lp = '\0';
+
+	type = DecodeUnits(0, lowunits, &val);
+	if (type == UNKNOWN_FIELD)
+		type = DecodeSpecial(0, lowunits, &val);
+
+	if (type == UNITS)
+	{
+		fsec_t		fsec;
+		struct tm	tt,
+				   *tm = &tt;
+
+		time2tm(time, tm, &fsec);
+
+		switch (val)
+		{
+			case DTK_MICROSEC:
+#ifdef HAVE_INT64_TIMESTAMP
+				result = ((tm->tm_sec * INT64CONST(1000000)) + fsec);
+#else
+				result = ((tm->tm_sec + fsec) * 1000000);
+#endif
+				break;
+
+			case DTK_MILLISEC:
+#ifdef HAVE_INT64_TIMESTAMP
+				result = ((tm->tm_sec * INT64CONST(1000))
+						  + (fsec / INT64CONST(1000)));
+#else
+				result = ((tm->tm_sec + fsec) * 1000);
+#endif
+				break;
+
+			case DTK_SECOND:
+#ifdef HAVE_INT64_TIMESTAMP
+				result = (tm->tm_sec + (fsec / INT64CONST(1000000)));
+#else
+				result = (tm->tm_sec + fsec);
+#endif
+				break;
+
+			case DTK_MINUTE:
+				result = tm->tm_min;
+				break;
+
+			case DTK_HOUR:
+				result = tm->tm_hour;
+				break;
+
+			case DTK_TZ:
+			case DTK_TZ_MINUTE:
+			case DTK_TZ_HOUR:
+			case DTK_DAY:
+			case DTK_MONTH:
+			case DTK_QUARTER:
+			case DTK_YEAR:
+			case DTK_DECADE:
+			case DTK_CENTURY:
+			case DTK_MILLENNIUM:
+			default:
+				elog(ERROR, "TIME units '%s' not supported",
+					 DatumGetCString(DirectFunctionCall1(textout,
+											   PointerGetDatum(units))));
+				result = 0;
+		}
+	}
+	else if ((type == RESERV) && (val == DTK_EPOCH))
+	{
+#ifdef HAVE_INT64_TIMESTAMP
+		result = (time / 1000000e0);
+#else
+		result = time;
+#endif
+	}
+	else
+	{
+		elog(ERROR, "TIME units '%s' not recognized",
+			 DatumGetCString(DirectFunctionCall1(textout,
+											   PointerGetDatum(units))));
+		result = 0;
+	}
+
+	PG_RETURN_FLOAT8(result);
+}
+
 
 /*****************************************************************************
  *	 Time With Time Zone ADT
  *****************************************************************************/
 
+/* tm2timetz()
+ * Convert a tm structure to a time data type.
+ */
+static int
+tm2timetz(struct tm * tm, fsec_t fsec, int tz, TimeTzADT *result)
+{
+#ifdef HAVE_INT64_TIMESTAMP
+	result->time = ((((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec)
+					 * INT64CONST(1000000)) + fsec);
+#else
+	result->time = ((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec + fsec);
+#endif
+	result->zone = tz;
+
+	return 0;
+}
+
 Datum
 timetz_in(PG_FUNCTION_ARGS)
 {
@@ -939,7 +1317,7 @@ timetz_in(PG_FUNCTION_ARGS)
 #endif
 	int32		typmod = PG_GETARG_INT32(2);
 	TimeTzADT  *result;
-	double		fsec;
+	fsec_t		fsec;
 	struct tm	tt,
 			   *tm = &tt;
 	int			tz;
@@ -950,18 +1328,15 @@ timetz_in(PG_FUNCTION_ARGS)
 	int			ftype[MAXDATEFIELDS];
 
 	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad time with time zone external representation"
-			 " (too long) '%s'", str);
+		elog(ERROR, "Bad time with time zone"
+			 " external representation (too long) '%s'", str);
 
 	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
 	  || (DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, &tz) != 0))
 		elog(ERROR, "Bad time external representation '%s'", str);
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
-
-	result->time = ((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec + fsec);
-	result->zone = tz;
-
+	tm2timetz(tm, fsec, tz, result);
 	AdjustTimeForTypmod(&(result->time), typmod);
 
 	PG_RETURN_TIMETZADT_P(result);
@@ -974,23 +1349,46 @@ timetz_out(PG_FUNCTION_ARGS)
 	char	   *result;
 	struct tm	tt,
 			   *tm = &tt;
-	double		fsec;
+	fsec_t		fsec;
 	int			tz;
-	double		trem;
 	char		buf[MAXDATELEN + 1];
 
+	timetz2tm(time, tm, &fsec, &tz);
+	EncodeTimeOnly(tm, fsec, &tz, DateStyle, buf);
+
+	result = pstrdup(buf);
+	PG_RETURN_CSTRING(result);
+}
+
+/* timetz2tm()
+ * Convert TIME WITH TIME ZONE data type to POSIX time structure.
+ * For dates within the system-supported time_t range, convert to the
+ *	local time zone. If out of this range, leave as GMT. - tgl 97/05/27
+ */
+static int
+timetz2tm(TimeTzADT *time, struct tm * tm, fsec_t *fsec, int *tzp)
+{
+#ifdef HAVE_INT64_TIMESTAMP
+	tm->tm_hour = (time->time / INT64CONST(3600000000));
+	time->time -= (tm->tm_hour * INT64CONST(3600000000));
+	tm->tm_min = (time->time / INT64CONST(60000000));
+	time->time -= (tm->tm_min * INT64CONST(60000000));
+	tm->tm_sec = (time->time / INT64CONST(1000000));
+	*fsec = (time->time - (tm->tm_sec * INT64CONST(1000000)));
+#else
+	double		trem;
+
 	trem = time->time;
 	TMODULO(trem, tm->tm_hour, 3600e0);
 	TMODULO(trem, tm->tm_min, 60e0);
 	TMODULO(trem, tm->tm_sec, 1e0);
-	fsec = trem;
+	*fsec = trem;
+#endif
 
-	tz = time->zone;
+	if (tzp != NULL)
+		*tzp = time->zone;
 
-	EncodeTimeOnly(tm, fsec, &tz, DateStyle, buf);
-
-	result = pstrdup(buf);
-	PG_RETURN_CSTRING(result);
+	return 0;
 }
 
 /* timetz_scale()
@@ -1118,7 +1516,7 @@ timetz_hash(PG_FUNCTION_ARGS)
 	 * sizeof(TimeTzADT), so that any garbage pad bytes in the structure
 	 * won't be included in the hash!
 	 */
-	return hash_any((char *) key, sizeof(double) + sizeof(int4));
+	return hash_any((unsigned char *) key, sizeof(key->time) + sizeof(key->zone));
 }
 
 Datum
@@ -1156,14 +1554,25 @@ timetz_pl_interval(PG_FUNCTION_ARGS)
 	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(0);
 	Interval   *span = PG_GETARG_INTERVAL_P(1);
 	TimeTzADT  *result;
+
+#ifndef HAVE_INT64_TIMESTAMP
 	TimeTzADT	time1;
+#endif
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 
+#ifdef HAVE_INT64_TIMESTAMP
+	result->time = (time->time + span->time);
+	result->time -= (result->time / INT64CONST(86400000000) * INT64CONST(86400000000));
+	if (result->time < INT64CONST(0))
+		result->time += INT64CONST(86400000000);
+#else
 	result->time = (time->time + span->time);
 	TMODULO(result->time, time1.time, 86400e0);
 	if (result->time < 0)
 		result->time += 86400;
+#endif
+
 	result->zone = time->zone;
 
 	PG_RETURN_TIMETZADT_P(result);
@@ -1178,14 +1587,25 @@ timetz_mi_interval(PG_FUNCTION_ARGS)
 	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(0);
 	Interval   *span = PG_GETARG_INTERVAL_P(1);
 	TimeTzADT  *result;
+
+#ifndef HAVE_INT64_TIMESTAMP
 	TimeTzADT	time1;
+#endif
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 
+#ifdef HAVE_INT64_TIMESTAMP
+	result->time = (time->time - span->time);
+	result->time -= (result->time / INT64CONST(86400000000) * INT64CONST(86400000000));
+	if (result->time < INT64CONST(0))
+		result->time += INT64CONST(86400000000);
+#else
 	result->time = (time->time - span->time);
 	TMODULO(result->time, time1.time, 86400e0);
 	if (result->time < 0)
 		result->time += 86400;
+#endif
+
 	result->zone = time->zone;
 
 	PG_RETURN_TIMETZADT_P(result);
@@ -1338,9 +1758,11 @@ time_timetz(PG_FUNCTION_ARGS)
 	TimeTzADT  *result;
 	struct tm	tt,
 			   *tm = &tt;
+	fsec_t		fsec;
 	int			tz;
 
-	GetCurrentTime(tm);
+	GetCurrentDateTime(tm);
+	time2tm(time, tm, &fsec);
 	tz = DetermineLocalTimeZone(tm);
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
@@ -1363,19 +1785,18 @@ timestamptz_timetz(PG_FUNCTION_ARGS)
 	struct tm	tt,
 			   *tm = &tt;
 	int			tz;
-	double		fsec;
+	fsec_t		fsec;
 	char	   *tzn;
 
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
 	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
-		elog(ERROR, "Unable to convert timestamp to date");
+		elog(ERROR, "Unable to convert timestamptz to timetz");
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 
-	result->time = ((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec + fsec);
-	result->zone = tz;
+	tm2timetz(tm, fsec, tz, result);
 
 	PG_RETURN_TIMETZADT_P(result);
 }
@@ -1394,7 +1815,12 @@ datetimetz_timestamptz(PG_FUNCTION_ARGS)
 	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(1);
 	TimestampTz result;
 
-	result = date * 86400.0 + time->time + time->zone;
+#ifdef HAVE_INT64_TIMESTAMP
+	result = (((date * INT64CONST(86400000000)) + time->time)
+			  + (time->zone * INT64CONST(1000000)));
+#else
+	result = (((date * 86400.0) + time->time) + time->zone);
+#endif
 
 	PG_RETURN_TIMESTAMP(result);
 }
@@ -1488,19 +1914,13 @@ timetz_part(PG_FUNCTION_ARGS)
 
 	if (type == UNITS)
 	{
-		double		trem;
 		double		dummy;
 		int			tz;
-		double		fsec;
+		fsec_t		fsec;
 		struct tm	tt,
 				   *tm = &tt;
 
-		trem = time->time;
-		TMODULO(trem, tm->tm_hour, 3600e0);
-		TMODULO(trem, tm->tm_min, 60e0);
-		TMODULO(trem, tm->tm_sec, 1e0);
-		fsec = trem;
-		tz = time->zone;
+		timetz2tm(time, tm, &fsec, &tz);
 
 		switch (val)
 		{
@@ -1519,15 +1939,28 @@ timetz_part(PG_FUNCTION_ARGS)
 				break;
 
 			case DTK_MICROSEC:
+#ifdef HAVE_INT64_TIMESTAMP
+				result = ((tm->tm_sec * INT64CONST(1000000)) + fsec);
+#else
 				result = ((tm->tm_sec + fsec) * 1000000);
+#endif
 				break;
 
 			case DTK_MILLISEC:
+#ifdef HAVE_INT64_TIMESTAMP
+				result = ((tm->tm_sec * INT64CONST(1000))
+						  + (fsec / INT64CONST(1000)));
+#else
 				result = ((tm->tm_sec + fsec) * 1000);
+#endif
 				break;
 
 			case DTK_SECOND:
+#ifdef HAVE_INT64_TIMESTAMP
+				result = (tm->tm_sec + (fsec / INT64CONST(1000000)));
+#else
 				result = (tm->tm_sec + fsec);
+#endif
 				break;
 
 			case DTK_MINUTE:
@@ -1553,7 +1986,13 @@ timetz_part(PG_FUNCTION_ARGS)
 		}
 	}
 	else if ((type == RESERV) && (val == DTK_EPOCH))
-		result = time->time + time->zone;
+	{
+#ifdef HAVE_INT64_TIMESTAMP
+		result = ((time->time / 1000000e0) - time->zone);
+#else
+		result = (time->time - time->zone);
+#endif
+	}
 	else
 	{
 		elog(ERROR, "TIMETZ units '%s' not recognized",
@@ -1574,7 +2013,6 @@ timetz_zone(PG_FUNCTION_ARGS)
 	text	   *zone = PG_GETARG_TEXT_P(0);
 	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(1);
 	TimeTzADT  *result;
-	TimeADT		time1;
 	int			tz;
 	int			type,
 				val;
@@ -1600,10 +2038,20 @@ timetz_zone(PG_FUNCTION_ARGS)
 	if ((type == TZ) || (type == DTZ))
 	{
 		tz = val * 60;
-		time1 = time->time - time->zone + tz;
-		TMODULO(result->time, time1, 86400e0);
-		if (result->time < 0)
+#ifdef HAVE_INT64_TIMESTAMP
+		result->time = time->time + ((time->zone - tz) * INT64CONST(1000000));
+		while (result->time < INT64CONST(0))
+			result->time += INT64CONST(86400000000);
+		while (result->time >= INT64CONST(86400000000))
+			result->time -= INT64CONST(86400000000);
+#else
+		result->time = time->time + (time->zone - tz);
+		while (result->time < 0)
 			result->time += 86400;
+		while (result->time >= 86400)
+			result->time -= 86400;
+#endif
+
 		result->zone = tz;
 	}
 	else
@@ -1631,15 +2079,27 @@ timetz_izone(PG_FUNCTION_ARGS)
 			 DatumGetCString(DirectFunctionCall1(interval_out,
 												 PointerGetDatum(zone))));
 
+#ifdef HAVE_INT64_TIMESTAMP
+	tz = -(zone->time / INT64CONST(1000000));
+#else
 	tz = -(zone->time);
+#endif
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 
-	result->time = (time->time + (time->zone - tz));
+#ifdef HAVE_INT64_TIMESTAMP
+	result->time = time->time + ((time->zone - tz) * INT64CONST(1000000));
+	while (result->time < INT64CONST(0))
+		result->time += INT64CONST(86400000000);
+	while (result->time >= INT64CONST(86400000000))
+		result->time -= INT64CONST(86400000000);
+#else
+	result->time = time->time + (time->zone - tz);
 	while (result->time < 0)
 		result->time += 86400;
 	while (result->time >= 86400)
 		result->time -= 86400;
+#endif
 
 	result->zone = tz;
 
